/*
 * Copyright (C) 1998,1999
 *	SHUDO Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

/*
 * ESI: vars
 * EDX: cache of stack top
 * ECX: cache of stack top
 */

/*
 * 5 states of Java stack
 *
 * state 0:
 *   ECX: undefined
 *   EDX: undefined
 * state 1:
 *   ECX: undefined
 *   EDX: stack top element
 * state 2:
 *   ECX: stack top element
 *   EDX: stack top-1 element
 * state 3:
 *   EDX: undefined
 *   ECX: stack top element
 * state 4:
 *   EDX: stack top element
 *   ECX: stack top-1 element
 */

#define	ST0	0
#define	ST1	1
#define	ST2	2
#define	ST3	3
#define	ST4	4
#define STANY	5
#define STSTA	5

#define NSTATES	5

#define STATE_AFTER_JUMP	ST2
	/* state for next instruction of goto,jsr,ret */


/*
 * OS dependence
 */

#if defined(linux) || (defined(sun) && defined(__svr4__))
#  define SYMBOL(SYM)	STR(SYM)
#elif defined(__FreeBSD__)
#  define SYMBOL(SYM)	"_" STR(SYM)
#endif

#ifndef WIN32
#  define	_drem	SYMBOL(fmod)
#endif


/*
 * space filled by compiler
 */

/* constant */
#define CONST	0x606060
/* for exception handler */
#define BYTEPCOFF	0x626262
#define ADDR_EXC	0x707070
/* for jump instructions */
#define ADDR_JP		0x727272
/* for *return instructions */
#define ADDR_FIN	0x747474


/*
 * utilities
 */

#ifdef GCC27
#define PUSH_CONSTSTR(STR)	__asm__("PUSH %0" : : "m" (STR))
#else
#define PUSH_CONSTSTR(STR)	__asm__("pushl %0" : : "m" (STR))
#endif

#if defined(linux)
#  define PUSH_STDOUT \
  __asm__("movl  _IO_stdout_@GOT(%ebx),%eax\n\t"\
	  "pushl %eax");
#elif defined(__FreeBSD__)
#  define PUSH_STDOUT \
  __asm__("movl  ___sF@GOT(%ebx),%eax\n\t"\
	  "leal  88(%eax),%eax\n\t"\
	  "pushl %eax");
#elif defined(sun) && defined(__svr4__)
#  define PUSH_STDOUT \
  __asm__("movl  __iob@GOT(%ebx),%eax\n\t"\
	  "leal  16(%eax),%eax\n\t"\
	  "pushl %eax");
#endif

#define FFLUSH \
  PUSH_STDOUT;\
  __asm__("call  " SYMBOL(fflush) "@PLT\n\t"\
	  "addl  $4,%esp")

#if 0
#define DEBUG_IN \
  __asm__("pushl %eax\n\tpushl %edx\n\tpushl %ebx\n\tpushl %ecx\n\t"\
	  "pushl %edi\n\tpushl %esi")
#define DEBUG_OUT \
  __asm__("popl  %esi\n\tpopl  %edi\n\t"\
	  "popl  %ecx\n\tpopl  %ebx\n\tpopl  %edx\n\tpopl  %eax")
#else
#define DEBUG_IN	__asm__("pusha")
#define DEBUG_OUT	__asm__("popa")
#endif

#define FUNCCALL_IN(STATE)	_FUNCCALL_IN(STATE)
#define FUNCCALL_OUT(STATE)	_FUNCCALL_OUT(STATE)
#define _FUNCCALL_IN(STATE)	FUNCCALL_IN_##STATE
#define _FUNCCALL_OUT(STATE)	FUNCCALL_OUT_##STATE
#define FUNCCALL_IN_0	__asm__("pushl %esi")
#define FUNCCALL_OUT_0	__asm__("popl  %esi")
#define FUNCCALL_IN_1	__asm__("pushl %esi\n\tpushl %edx")
#define FUNCCALL_OUT_1	__asm__("popl  %edx\n\tpopl  %esi")
#define FUNCCALL_IN_2	__asm__("pushl %esi\n\tpushl %edx\n\tpushl %ecx")
#define FUNCCALL_OUT_2	__asm__("popl  %ecx\n\tpopl  %edx\n\tpopl  %esi")
#define FUNCCALL_IN_3	__asm__("pushl %esi\n\tpushl %ecx")
#define FUNCCALL_OUT_3	__asm__("popl  %ecx\n\tpopl  %esi")
#define FUNCCALL_IN_4	__asm__("pushl %esi\n\tpushl %ecx\n\tpushl %edx")
#define FUNCCALL_OUT_4	__asm__("popl  %edx\n\tpopl  %ecx\n\tpopl  %esi")


#define STATETO00
#define STATETO10	__asm__("pushl %edx")
#define STATETO20	__asm__("pushl %edx\n\tpushl %ecx")
#define STATETO30	__asm__("pushl %ecx")
#define STATETO40	__asm__("pushl %ecx\n\tpushl %edx")

#define SIGNAL_ERROR_JUMP(STATE) \
  bytepcoff = 0x626262;\
  STATETO##STATE##0;\
  __asm__("jmp   " STR(ADDR_EXC))

#define SIGNAL_ERROR_CORE(KIND, STATE) \
  PUSH_CONSTSTR(JAVAPKG #KIND);		/* char *ename */\
  __asm__("pushl %0" : : "m" (ee));	/* ExecEnv *ee */\
  __asm__("call  " SYMBOL(SignalError) "@PLT\n\t"\
	  "addl  $12,%esp");\
  SIGNAL_ERROR_JUMP(STATE)

#define SIGNAL_ERROR0(KIND, STATE) \
  __asm__("pushl $0");			/* char *DetailMessage */\
  SIGNAL_ERROR_CORE(KIND, STATE)

#define SIGNAL_ERROR1(KIND, MSG, STATE) \
  PUSH_CONSTSTR(MSG);			/* char *DetailMessage */\
  SIGNAL_ERROR_CORE(KIND, STATE)


#ifndef NO_CHECK
#  define NULL_TEST(REG, LABEL, STATE) \
  __asm__("testl " #REG "," #REG "\n\t"\
	  "jnz   " LABEL);\
  SIGNAL_ERROR0(NullPointerException, STATE);\
  __asm__(LABEL ":")
#else
#  define NULL_TEST(REG, LABEL)
#endif


/*
 * code header
 *   for each opcode of JVM.
 *
 * 0x0f, 0x0f,
 * 0xe9, opcode(2 byte), initial state(1 byte), last state(1 byte)
 */

#ifdef RUNTIME_DEBUG
#  define CODE_DEBUG(LABEL) \
    __asm__("pushl %eax\n\t"\
	    "movl  %esp,%eax");\
  DEBUG_IN;\
  if (runtime_debug) {\
    __asm__("addl  $4,%eax\n\t"\
	    "pushl %eax");	/* %esp */\
    PUSH_CONSTSTR(">" LABEL " %x\n");\
    __asm__("call " SYMBOL(printf) "@PLT\n\t"\
	    "addl $8,%esp");\
    FFLUSH;\
  }\
  DEBUG_OUT;\
  __asm__("popl  %eax");
#else
#  define CODE_DEBUG(LABEL)
#endif

#define CODE(OPCODE, LABEL, STATE, NEXTST) \
  _CODE(OPCODE, LABEL, STATE, NEXTST);
#define	_CODE(OPCODE, LABEL, STATE, NEXTST) \
  __asm__(\
  ".byte 0x0f,0x0f\n\t"\
  ".byte 0xe9," #OPCODE ",0," #STATE "," #NEXTST);\
  CODE_DEBUG(#LABEL "(" #STATE "->" #NEXTST ")")

#define CODEEND	__asm__(".byte 0x0f,0x0f,0x0f,0x0f")


/*
 * JDK related stuffs
 */

#define UNHAND(HANDLE, DST)	__asm__("movl  (" #HANDLE ")," #DST)

#define OBJ_MONITOR(HANDLE, DST)
#define OBJ_FLAGS(HANDLE, DST) \
  __asm__("movl  4(" #HANDLE ")," #DST "\n\t"\
	  "andl  $31," #DST)
#define OBJ_LENGTH(HANDLE, DST) \
  __asm__("movl  4(" #HANDLE ")," #DST "\n\t"\
	  "shrl  $5," #DST)
#define OBJ_METHODTABLE(HANDLE, DST) \
  __asm__("movl  4(" #HANDLE ")," #DST)
#define OBJ_ARRAY_METHODTABLE(HANDLE, DST, LABEL) \
  __asm__("movl  4(" #HANDLE ")," #DST "\n\t"\
	  "movl  " #DST ",%eax\n\t"\
	  "andl  $31,%eax\n\t"\
	  "cmpl  $" STR(T_NORMAL_OBJECT) ",%eax\n\t"\
	  "je    " LABEL "_mtdone");\
  __asm__("movl  %0,%" #DST : : "m" (classJavaLangObject));\
  __asm__("movl  (" #DST "),%eax\n\t"\
	  "movl  48(%eax)," #DST);\
  __asm__(LABEL "_mtdone:")
	/* destroy %eax */

#define MT_SLOT(MTBL, SLOT, DST) \
  __asm__("movl  4(" #MTBL "," #SLOT ",4)," #DST)
#define MT_CLASSDESCRIPTOR(MTBL, DST) \
  __asm__("movl  (" #MTBL ")," #DST)

#define UOBJ_GETSLOT(OBJ, SLOT, DST) \
  __asm__("movl  (" #OBJ "," #SLOT ",4)," #DST)

#define OBJ_GETSLOT(HANDLE, SLOT, DST) \
  UNHAND(HANDLE, %edi);\
  UOBJ_GETSLOT(%edi, SLOT, DST)

#define UOBJ_GETSLOT2(OBJ, SLOT, DST_LOW, DST_HIGH) \
  __asm__("leal  (" #OBJ "," #SLOT ",4),%edi\n\t"\
	  "movl  (%edi)," #DST_LOW "\n\t"\
	  "movl  4(%edi)," #DST_HIGH)

#define OBJ_GETSLOT2(HANDLE, SLOT, DST_LOW, DST_HIGH) \
  UNHAND(HANDLE, %edi);\
  UOBJ_GETSLOT2(%edi, SLOT, DST_LOW, DST_HIGH)

#define UOBJ_SETSLOT(OBJ, SLOT, VAL) \
  __asm__("movl  " #VAL ",(" #OBJ "," #SLOT ",4)")

#define OBJ_SETSLOT(HANDLE, SLOT, VAL) \
  UNHAND(HANDLE, %edi);\
  UOBJ_SETSLOT(%edi, SLOT, VAL)

#define UOBJ_SETSLOT2(OBJ, SLOT, VAL_LOW, VAL_HIGH) \
  __asm__("leal  (" #OBJ "," #SLOT ",4),%edi\n\t"\
	  "movl  " #VAL_LOW ",(%edi)\n\t"\
	  "movl  " #VAL_HIGH ",4(%edi)")

#define OBJ_SETSLOT2(HANDLE, SLOT, VAL_LOW, VAL_HIGH) \
  UNHAND(HANDLE, %edi);\
  UOBJ_SETSLOT2(%edi, SLOT, VAL_LOW, VAL_HIGH)

#define CB_NAME(CLAZZ, DST) \
  __asm__("movl  (" #CLAZZ "),%edi\n\t"\
	  "movl  4(%edi)," #DST)
#define CB_ACCESS(CLAZZ, DST) \
  __asm__("movl  (" #CLAZZ "),%edi\n\t"\
	  "movzwl 86(%edi)," #DST)
#define CB_METHODTABLE(CLAZZ, DST) \
  __asm__("movl  (" #CLAZZ "),%edi\n\t"\
	  "movl  48(%edi)," #DST)

#define EE_CURRENTFRAME(EE)	"4(" #EE ")"
#define EE_EXCEPTIONKIND(EE)	"12(" #EE ")"

#define JUMP_IF_EXC_HASNT_OCCURRED(EE, LABEL) \
  __asm__("movl  " EE_EXCEPTIONKIND(EE) ",%edi\n\t"\
	  "testl %edi,%edi\n\t"\
	  "jz    " LABEL)
#define EXCEPTION_CLEAR(EE) \
  __asm__("movl  $0," EE_EXCEPTIONKIND(EE));

#define FRAME_OPTOP(FRAME)	"8(" #FRAME ")"
#define FRAME_VARS(FRAME)	"12(" #FRAME ")"
#define FRAME_PREV(FRAME)	"16(" #FRAME ")"
#define FRAME_LASTPC(FRAME)	"24(" #FRAME ")"
#define FRAME_CURRENTMETHOD(FRAME)	"28(" #FRAME ")"
#define FRAME_OSTACK(FRAME)	"40(" #FRAME ")"



#define METHOD_NLOCALS(MB, DST) \
  __asm__("movzwl 64(" #MB ")," #DST)
#define METHOD_CLAZZ(MB, DST) \
  __asm__("movl  (" #MB ")," #DST)
#define METHOD_SIGNATURE(MB)	"4(" #MB ")"
#define METHOD_NAME(MB)	"8(" #MB ")"


#ifdef JITDO
/*
 * JITDO related stuff
 */
#define EE_REMOTE_FLAG(EE)	"61(" #EE ")"
	/* ee->alloc_cache.cache_pad[0] */
#define EE_REMOTE_ADDR(EE)	"16(" #EE ")"
	/* ee->exception.exc */

#define JUMP_IF_NOT_REMOTE(LABEL) \
    {	/* %edi = ee */\
      register ExecEnv *cur_ee asm("edi");\
      cur_ee = ee;\
    }\
    __asm__("movsbl " EE_REMOTE_FLAG(%edi) ",%edi");\
    __asm__("testl %edi,%edi\n\t"\
	    "jz    " LABEL)
	/* break %edi */

#define METHODTABLE_OF_PROXY(DST) \
    __asm__("movl  " SYMBOL(proxy_methodtable) "@GOT(%ebx)," #DST "\n\t"\
	    "movl  (" #DST ")," #DST);
	/* break no registers */

#define JUMP_IF_NOT_PROXY(HANDLE_REG /* must not be %edi */, LABEL) \
    METHODTABLE_OF_PROXY(%edi);\
    __asm__("cmpl  4(" #HANDLE_REG "),%edi\n\t"\
	    "jnz   " LABEL)
	/* break %edi */

#define PROXY_CLAZZ(HANDLE, DST) \
    UNHAND(HANDLE, DST);\
    __asm__("movl  4(" #DST ")," #DST);

#endif	/* JITDO */
